
 /********************************************************************\
 *                                                                    *
 * Netscape plugin to talk to festival to do TTS.                     *
 *                                                                    *
 \********************************************************************/

#include <stdio.h>
#include <stdarg.h>

#include "npapi.h"
#include "netscape_plugin_Plugin.h"

#include <sys/types.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/xpm.h>

#define IMPLEMENT_FestivalPlugin
#include "FestivalPlugin.h"

JRIEnv *java_env;
struct java_lang_Class *java_class;

static struct protoent *tcp_protocol;
static struct sockaddr_in festival_address;
static char audioserver[255];
static Pixmap logo=None;
static Display *logo_display;
static int logo_width, logo_height;

static int connect_to_festival(void)
{
  int s;

  if ((s=socket(PF_INET, SOCK_STREAM, tcp_protocol->p_proto)) < 0)
	{
	  fprintf(stderr, "can't create socket\n");
	  return -3;
	}

  if (connect(s, (struct sockaddr *)&festival_address, sizeof(festival_address)) < 0)
	{
	  close(s);
	  fprintf(stderr, "can't connect to %s:%d\n", 
			  inet_ntoa(festival_address.sin_addr), festival_address.sin_port);
	  return -4;
	}

  return s;
}

static int lookup_server(const char *host, int port)
{
  struct hostent *hostentp;
  int s;
  
  if (!(tcp_protocol = getprotobyname("tcp")))
	{
	  fprintf(stderr, "can't find TCP protocol\n");
	  return 1;
	}

  if ((hostentp=gethostbyname(host))==NULL)
	{
	  fprintf(stderr, "can't find host '%s'\n", host);
	  return 2;
	}

  /* There has to be a better way. This is how I've always done it,
   * but somewhere there must be a better way somewhere in
   * the library! */

  bzero(&(festival_address.sin_addr) ,sizeof(struct in_addr));
  festival_address.sin_family=hostentp->h_addrtype;
  festival_address.sin_port=port;
  memmove(&festival_address.sin_addr, 
		  (hostentp->h_addr_list)[0],
		  hostentp->h_length);

  if ((s=connect_to_festival()) < 0)
	return s;

  close(s);

  return 0;
}

static void printf_to_festival(int s, const char *format, void *arg)
{
  char buffer[1024];

  sprintf(buffer, format, arg);

  fprintf(stderr, "send: '%s'\n", buffer);

  write(s, buffer, strlen(buffer));
}

static void Redraw(Widget w, XtPointer closure, XEvent *event);
static Pixmap buildLogo(Display *dpy, Window w, int *widthp, int *heightp, Pixmap *maskp);

 /********************************************************************\
 *                                                                    *
 * Information for one particuar festival session.                    *
 *                                                                    *
 \********************************************************************/

typedef struct _PluginInstance
{
  int socket;
  Window window;
  Display *display;
  uint32 x, y;
  uint32 width, height;
} PluginInstance;


 /********************************************************************\
 *                                                                    *
 * How should this work?                                              *
 *                                                                    *
 \********************************************************************/

char*
NPP_GetMIMEDescription(void)
{
	return("application/x-tts:tts:Festival Plugin");
}

NPError
NPP_GetValue(void *future, NPPVariable variable, void *value)
{
	NPError err = NPERR_NO_ERROR;

	switch (variable) {
		case NPPVpluginNameString:
			*((char **)value) = "Festival plugin";
			break;
		case NPPVpluginDescriptionString:
			*((char **)value) =
				"A plugin allowing javascript to talk to festival.";
			break;
		default:
			err = NPERR_GENERIC_ERROR;
	}
	return err;
}

NPError
NPP_Initialize(void)
{
  const char *server = getenv("FESTIVAL");
  const char *audio = getenv("AUDIOSERVER");
  char hostname[255];
  int port;

  if (audio)
	fprintf(stderr, "AUDIOSERVER = '%s'\n", audio);
  
  if (!audio)
	if (!(audio = getenv("DISPLAY")))
	  {
		fprintf(stderr, "Neither AUDIOSERVER nor DISPLAY set\n");
		return NPERR_INVALID_PARAM;
	  }

  if (audio[0] == ':')
	{
	  hostname[0] ='\0';
	  port=atoi(audio+1);
	}
  else
	if (sscanf(audio, "%[^:]:%d", hostname, &port) != 2)
	  {
		fprintf(stderr, "'%s' not in format host:port\n", audio);
		return NPERR_INVALID_PARAM;
	  }

  if (hostname[0] == '\0' || strcmp(hostname, "localhost")==0)
	gethostname(hostname, 255);

  sprintf(audioserver, "%s:%d", hostname, port);

  fprintf(stderr, "audio server = %s\n", audioserver);

  if (!server)
	{
	  fprintf(stderr, "Environment variable FESTIVAL not set\n");
	  return NPERR_INVALID_PARAM;
	}

  if (sscanf(server, "%[^:]:%d", hostname, &port) != 2)
	{
	  fprintf(stderr, "'%s' not in format host:port\n", server);
	  return NPERR_INVALID_PARAM;
	}
  
  if (lookup_server(hostname, port) != 0)
	return NPERR_INVALID_PARAM;

  return NPERR_NO_ERROR;
}


jref
NPP_GetJavaClass()
{
  java_env = NPN_GetJavaEnv();

  if (!java_env)
	{
	  fprintf(stderr, "java disabled\n");
	  return NULL;
	}

  if (!(java_class = use_FestivalPlugin(java_env)))
	{
	  fprintf(stderr, "can't load FestivalPlugin class\n");
	  java_env=NULL;
	}
  else
	fprintf(stderr, "loaded FestivalPlugin class\n");

  return java_class;
}

void
NPP_Shutdown(void)
{
  if (java_env)
	unuse_FestivalPlugin(java_env);

  if (logo != None)
	{
	XFreePixmap(logo_display, logo);
	logo=None;
	}
}

NPError 
NPP_New(NPMIMEType pluginType,
	NPP instance,
	uint16 mode,
	int16 argc,
	char* argn[],
	char* argv[],
	NPSavedData* saved)
{
        PluginInstance* This;

	if (instance == NULL)
		return NPERR_INVALID_INSTANCE_ERROR;
		
	instance->pdata = NPN_MemAlloc(sizeof(PluginInstance));
	
	This = (PluginInstance*) instance->pdata;

	if (This == NULL)
		return NPERR_OUT_OF_MEMORY_ERROR;

	if ((This->socket = connect_to_festival()) < 0)
	  return NPERR_GENERIC_ERROR;

	printf_to_festival(This->socket, "(Parameter.set 'Audio_Method 'Audio_Command)\n",NULL);
	printf_to_festival(This->socket, "(Parameter.set 'Audio_Command \"na_play -server %s -itype raw -f $SR $FILE\")\n", audioserver);
	printf_to_festival(This->socket, "(print (Parameter.get 'Audio_Command))\n",NULL);


	return NPERR_NO_ERROR;
}


NPError 
NPP_Destroy(NPP instance, NPSavedData** save)
{
	PluginInstance* This;

	if (instance == NULL)
		return NPERR_INVALID_INSTANCE_ERROR;

	This = (PluginInstance*) instance->pdata;

	if (This->socket >=0)
	  if (close(This->socket) != 0)
		  fprintf(stderr, "can't close connection to festival\n");

	if (This != NULL) {
		NPN_MemFree(instance->pdata);
		instance->pdata = NULL;
	}

	return NPERR_NO_ERROR;
}



NPError 
NPP_SetWindow(NPP instance, NPWindow* window)
{
	PluginInstance* This;

	if (instance == NULL)
		return NPERR_INVALID_INSTANCE_ERROR;

	if (window == NULL)
		return NPERR_NO_ERROR;

	This = (PluginInstance*) instance->pdata;

	{
	  Widget netscape_widget;

	  This->window = (Window) window->window;
	  This->x = window->x;
	  This->y = window->y;
	  This->width = window->width;
	  This->height = window->height;
	  This->display = ((NPSetWindowCallbackStruct *)window->ws_info)->display;

	  if (logo==None)
		{
		logo = buildLogo(This->display, This->window, &logo_width, &logo_height, NULL);
		logo_display = This->display;
		}
		
	  netscape_widget = XtWindowToWidget(This->display, This->window);
	  XtAddEventHandler(netscape_widget, ExposureMask, FALSE, (XtEventHandler)Redraw, This);
	  Redraw(netscape_widget, (XtPointer)This, NULL);
	}

	return NPERR_NO_ERROR;
}


NPError 
NPP_NewStream(NPP instance,
			  NPMIMEType type,
			  NPStream *stream, 
			  NPBool seekable,
			  uint16 *stype)
{
	NPByteRange range;
	PluginInstance* This;

	if (instance == NULL)
		return NPERR_INVALID_INSTANCE_ERROR;

	This = (PluginInstance*) instance->pdata;

	return NPERR_NO_ERROR;
}


/* Say we will accept any amount of input */

int32 STREAMBUFSIZE = 0X0FFFFFFF; 

int32 
NPP_WriteReady(NPP instance, NPStream *stream)
{
	PluginInstance* This;
	if (instance != NULL)
		This = (PluginInstance*) instance->pdata;

	return STREAMBUFSIZE;
}


int32 
NPP_Write(NPP instance, NPStream *stream, int32 offset, int32 len, void *buffer)
{
	if (instance != NULL)
	{
		PluginInstance* This = (PluginInstance*) instance->pdata;
	}

	return -1;		/* Any data is bad data */
}


NPError 
NPP_DestroyStream(NPP instance, NPStream *stream, NPError reason)
{
	PluginInstance* This;

	if (instance == NULL)
		return NPERR_INVALID_INSTANCE_ERROR;
	This = (PluginInstance*) instance->pdata;

	return NPERR_NO_ERROR;
}


void 
NPP_StreamAsFile(NPP instance, NPStream *stream, const char* fname)
{
	PluginInstance* This;
	if (instance != NULL)
		This = (PluginInstance*) instance->pdata;
}


void 
NPP_Print(NPP instance, NPPrint* printInfo)
{
	if(printInfo == NULL)
		return;

	if (instance != NULL) {
		PluginInstance* This = (PluginInstance*) instance->pdata;
	
		if (printInfo->mode == NP_FULL) {
			void* platformPrint =
				printInfo->print.fullPrint.platformPrint;
			NPBool printOne =
				printInfo->print.fullPrint.printOne;
			
			/* Do the default*/
			printInfo->print.fullPrint.pluginPrinted = FALSE;
		}
		else {	/* If not fullscreen, we must be embedded */
			NPWindow* printWindow =
				&(printInfo->print.embedPrint.window);
			void* platformPrint =
				printInfo->print.embedPrint.platformPrint;
		}
	}
}

 /********************************************************************\
 *                                                                    *
 * OK, here are the entry points from Java.                           *
 *                                                                    *
 \********************************************************************/

void
native_FestivalPlugin_say(JRIEnv* env,
						  FestivalPlugin *self,
						  struct java_lang_String *java_text)
{
  const char *text = JRI_GetStringUTFChars(env, java_text);
  NPP instance = (NPP)netscape_plugin_Plugin_getPeer(env, self);
  PluginInstance* This = (PluginInstance*) instance->pdata;
    
  fprintf(stderr, "tts: '%s'\n", text);

  printf_to_festival(This->socket,
					 "(SayText \"%s\")\n",
					 (void *)text);
}

void
native_FestivalPlugin_feval(JRIEnv* env,
						  FestivalPlugin *self,
						  struct java_lang_String *java_sexp)
{
  const char *sexp = JRI_GetStringUTFChars(env, java_sexp);
  NPP instance = (NPP)netscape_plugin_Plugin_getPeer(env, self);
  PluginInstance* This = (PluginInstance*) instance->pdata;
    
  fprintf(stderr, "eval: '%s'\n", sexp);

  printf_to_festival(This->socket,
					 "%s\n",
					 (void *)sexp);
}

 /********************************************************************\
 *                                                                    *
 * Some X11 stuff to put up the little icon.                          *
 *                                                                    *
 \********************************************************************/

static void
Redraw(Widget w, XtPointer closure, XEvent *event)
{
	PluginInstance* This = (PluginInstance*)closure;
	GC gc;
	const char* text = "FESTIVAL";
	XGCValues gcv;
#if 0
	Pixmap logo;
	int logo_width, logo_height;

	logo=buildLogo(This->display, This->window, &logo_width, &logo_height, NULL);
#endif

	XtVaGetValues(w, XtNbackground, &gcv.background,
				  XtNforeground, &gcv.foreground, 0);
	gcv.plane_mask = AllPlanes;
	gcv.function = GXcopy;
	gc = XCreateGC(This->display, This->window, 
				   GCForeground|GCBackground|GCPlaneMask|GCFunction, &gcv);

	if (logo == None)
	  {
		XDrawRectangle(This->display, This->window, gc, 
					   0, 0, This->width-1, This->height-1);
		XDrawString(This->display, This->window, gc, 
					This->width/2 - 22, This->height/2+5,
					text, strlen(text));
	  }
	else
	  {
		/* fprintf(stderr, "draw %x 0 0 %d %d\n", logo, logo_width, logo_height); */
#if 1
		XCopyArea(This->display, logo, This->window,
				  gc, 0, 0,
				  logo_width, logo_height,
				  0,0);
#else
		XCopyPlane(This->display, logo, This->window,
				   gc, 0, 0,
				   logo_width, logo_height,
				   0,0,1);
		XDrawRectangle(This->display, This->window, gc, 
					   0, 0,logo_width-1, logo_height-1);
#endif
	  }
}

static Pixmap buildLogo(Display *dpy, Window w, int *widthp, int *heightp, Pixmap *maskp)
{
  Pixmap pixmap=None;

#   include "logo.xpm"

  int XPMret;
  XpmColorSymbol bgcolor[2];
  XpmAttributes attribs;

  XWindowAttributes wat;

  XGetWindowAttributes(dpy, w, &wat);

    fprintf(stderr, "window = %d, %d, %d\n", wat.width, wat.height, wat.depth);

  attribs.valuemask = XpmExactColors|XpmColormap|XpmVisual|XpmDepth|XpmCloseness;
  attribs.colorsymbols = bgcolor;
  attribs.colormap = wat.colormap;
  attribs.visual = wat.visual;
  attribs.depth = wat.depth;
  attribs.numsymbols = 0;
  attribs.exactColors = 0;
  attribs.closeness = 16192;
  if ((XPMret = XpmCreatePixmapFromData(dpy, w,
										logo, &pixmap, maskp, &attribs)) >= 0) 
	{
	  *widthp=attribs.width;
	  *heightp=attribs.height;
	} 
  else 
	fprintf(stderr, "Error creating logo (Error code: %d)\n", XPMret); 

  return pixmap;
}


